import abc
import logging
import os
from typing import Dict, Any, Callable, Tuple

from google import auth
from google.auth.credentials import Credentials
from google.auth.impersonated_credentials import Credentials as ImpersonatedCredentials
from googleapiclient.discovery import build

DEFAULT_PROJECT_ID = os.environ.get('PROJECT_ID')
DEFAULT_LOCATION = os.environ.get('LOCATION', None)

PolicyType = Dict[str, Any]
AuthType = Tuple[Credentials, str]
ImpersonatedAuthType = Tuple[ImpersonatedCredentials, str]

logger = logging.getLogger()


def _get_project_default_location(credentials, project_id: str, default_zone: str = '1') -> str:
    service = build(serviceName='appengine', version='v1', credentials=credentials, cache_discovery=False)
    data = service.apps().get(appsId=project_id).execute()
    return data['locationId'] + default_zone


MINIMAL_SCOPES = [
    'https://www.googleapis.com/auth/cloud-platform',
]


class GoogleCloudPilotAPI(abc.ABC):
    _client_class = None
    _scopes = []
    _iam_roles = []
    _cached_credentials = None

    def __init__(
            self,
            subject: str = None,
            location: str = None,
            project_id: str = None,
            impersonate_account: str = None,
            **kwargs,
    ):
        self.credentials, credential_project_id = self._set_credentials(
            subject=subject,
            impersonate_account=impersonate_account,
        )
        self.project_id = self._set_project_id(project_id=project_id, credential_project_id=credential_project_id)
        self.location = self._set_location(location=location)

        self.client = (self._client_class or build)(
            credentials=self.credentials,
            **kwargs
        )

    def _set_project_id(self, project_id: str, credential_project_id: str) -> str:
        return project_id or DEFAULT_PROJECT_ID or credential_project_id

    def _set_location(self, location: str = None) -> str:
        return location or DEFAULT_LOCATION or _get_project_default_location(
            credentials=self.credentials,
            project_id=self.project_id,
        )

    @classmethod
    def _impersonate_account(cls, credentials, service_account, scopes) -> ImpersonatedAuthType:
        credentials = ImpersonatedCredentials(
            source_credentials=credentials,
            target_principal=service_account,
            target_scopes=scopes,
            delegates=[],
        )

        # Fetch project from service account
        # Since we are impersonating this service account, use it's own project
        # TODO: regex this
        project_id = credentials.service_account_email.split('@')[-1].replace('.iam.gserviceaccount.com', '')

        return credentials, project_id

    @classmethod
    def _set_credentials(cls, subject: str = None, impersonate_account: str = None) -> AuthType:
        # Speed up consecutive authentications
        # TODO: check if this does not break multiple client usage
        if not cls._cached_credentials:
            all_scopes = MINIMAL_SCOPES + cls._scopes
            credentials, project_id = auth.default(scopes=all_scopes)

            if impersonate_account:
                credentials, impersonated_project_id = cls._impersonate_account(
                    credentials=credentials,
                    service_account=impersonate_account,
                    scopes=all_scopes,
                )
                project_id = impersonated_project_id or project_id

            cls._cached_credentials = credentials, project_id

        credentials, project_id = cls._cached_credentials
        if subject:
            credentials = credentials.with_subject(subject=subject)
        return credentials, (project_id or getattr(credentials, 'project_id'))

    @property
    def oidc_token(self) -> Dict[str, Dict[str, str]]:
        return {'oidc_token': {'service_account_email': self.credentials.service_account_email}}

    async def add_permissions(self, email: str, project_id: str = None) -> None:
        from gcp_pilot.resource import GoogleResourceManager  # pylint: disable=import-outside-toplevel

        for role in self._iam_roles:
            await GoogleResourceManager().add_member(
                email=email,
                role=role,
                project_id=project_id or self.project_id,
            )

    def _get_project_number(self, project_id: str) -> int:
        from gcp_pilot.resource import GoogleResourceManager  # pylint: disable=import-outside-toplevel

        project = GoogleResourceManager().get_project(project_id=project_id)
        return project.projectNumber

    def _paginate(self, method: Callable, result_key: str, params: Dict[str, Any] = None) -> Dict[str, Any]:
        page_token = None
        params = params or {}

        while True:
            results = method(
                **params,
                pageToken=page_token,
            ).execute()
            for item in results.get(result_key, []):
                yield item

            page_token = results.get('nextPageToken')
            if not page_token:
                break


class AccountManagerMixin:
    def as_member(self, email: str) -> str:
        is_service_account = email.endswith('.gserviceaccount.com')
        prefix = 'serviceAccount' if is_service_account else 'member'
        return f'{prefix}:{email}'

    def bind_email_to_policy(self, email: str, role: str, policy: Dict) -> Dict:
        role_id = role if (role.startswith('organizations/') or role.startswith('roles/')) else f'roles/{role}'
        member = self.as_member(email=email)

        try:
            binding = next(b for b in policy['bindings'] if b['role'] == role_id)
            if member in binding['members']:
                return policy
            binding['members'].append(member)
        except (StopIteration, KeyError):
            binding = {'role': role_id, 'members': [member]}
            policy['bindings'] = policy.get('bindings', []).append(binding)
        return policy

    def unbind_email_from_policy(self, email: str, role: str, policy: Dict):
        role_id = f'roles/{role}'
        member = self.as_member(email=email)

        try:
            binding = next(b for b in policy['bindings'] if b['role'] == role_id)
            if member not in binding['members']:
                return policy
            binding['members'].remove(member)
        except (StopIteration, KeyError):
            pass
        return policy


class AppEngineBasedService:
    # This mixin must be used for all clients that uses GCP's resources that are based on App Engine
    # such as Cloud Scheduler.
    # They have a peculiar behaviour because the App Engine, when enabled the first time,
    # is created in a region and this region cannot be changed ever.
    # So these clients cannot just choose a desired region to work on, they must use the App Engine's
    # previously chosen region.
    def _set_location(self, location: str = None):
        project_location = _get_project_default_location(credentials=self.credentials, project_id=self.project_id)

        explicit_location = location
        if explicit_location and explicit_location != project_location:
            logger.warning(
                f"Location {explicit_location} cannot be set in {self.__class__.__name__}. "
                f"It uses App Engine's default location for your project: {project_location}"
            )
        return project_location



import datetime

import pytz

from gcp_pilot.base import GoogleCloudPilotAPI


class Calendar(GoogleCloudPilotAPI):
    _scopes = ['https://www.googleapis.com/auth/calendar']

    def __init__(self, email, timezone: str = 'UTC', **kwargs):
        self.email = email
        self.timezone = pytz.timezone(timezone)

        super().__init__(
            serviceName='calendar',
            version='v3',
            cache_discovery=False,
            subject=email,
            **kwargs,
        )

    class Color:
        light_blue = ('1', '#a4bdfc')
        light_green = ('2', '#7ae7bf')
        light_purple = ('3', '#dbadff')
        light_red = ('4', '#ff887c')
        yellow = ('5', '#fbd75b')
        light_orange = ('6', '#ffb878')
        water_green = ('7', '#46d6db')
        gray = ('8', '#e1e1e1')
        blue = ('9', '#5484ed')
        green = ('10', '#51b749')
        red = ('11', '#dc2127')

    def date_to_str(self, dt, fmt='%Y-%m-%dT%H:%M:%SZ'):
        if isinstance(dt, datetime.datetime):
            if not dt.tzinfo:
                dt = self.timezone.localize(dt)
            dt = dt.astimezone(self.timezone)
        elif isinstance(dt, datetime.date):
            dt = datetime.datetime(year=dt.year, month=dt.month, day=dt.day, tzinfo=self.timezone)

        return dt.strftime(fmt)

    def get_events(self, calendar_id='primary', starts_at=None):
        min_date = self.date_to_str(starts_at) if starts_at else None

        page_size = 100
        filters = dict(
            calendarId=calendar_id,
            timeMin=min_date,
            maxResults=page_size,
            singleEvents=True,
            orderBy='startTime',
        )
        page_token = None
        while True:
            filters['pageToken'] = page_token
            events_result = self.client.events().list(**filters).execute()

            events = events_result.get('items', [])
            for event in events:
                yield event

            page_token = events_result.get('nextPageToken')
            if not page_token:
                break

    def get_calendars(self):
        page_token = None
        while True:
            calendar_list = self.client.calendarList().list(pageToken=page_token).execute()
            for calendar_list_entry in calendar_list['items']:
                yield calendar_list_entry
            page_token = calendar_list.get('nextPageToken')
            if not page_token:
                break

    def create_or_update_event(
            self,
            summary,
            location,
            start_at,
            end_at,
            event_id=None,
            description=None,
            attendees=None,
            recurrence_type=None,
            recurrence_amount=None,
            calendar_id='primary',
            color=None,
    ):
        def _build_time_field(dt):
            if isinstance(start_at, datetime.datetime):
                return {
                    'dateTime': self.date_to_str(dt),
                }
            else:
                return {
                    'date': self.date_to_str(dt, fmt='%Y-%m-%d'),
                }

        data = {
            'summary': summary,
            'location': location,
            'description': description,
            'start': _build_time_field(dt=start_at),
            'end': _build_time_field(dt=end_at),
        }

        if color:
            data['colorId'] = color[0]

        if recurrence_type and recurrence_amount:
            data['recurrence'] = [
                f'RRULE:FREQ={recurrence_type};COUNT={recurrence_amount}'
            ]

        if attendees:
            data['attendees'] = attendees

        if event_id:
            return self.client.events().update(
                calendarId=calendar_id,
                eventId=event_id,
                body=data,
            ).execute()
        else:
            return self.client.events().insert(
                calendarId=calendar_id,
                body=data,
            ).execute()

    def delete_event(self, event_id, calendar_id='primary'):
        return self.client.events().delete(
            calendarId=calendar_id,
            eventId=event_id,
        ).execute()



from datetime import datetime
from typing import Any

from google.cloud import bigquery
from google.cloud.bigquery import Table, DatasetReference

from gcp_pilot import exceptions
from gcp_pilot.base import GoogleCloudPilotAPI


class BigQuery(GoogleCloudPilotAPI):
    _client_class = bigquery.Client

    def _dataset_ref(self, dataset_name: str, project_id: str = None) -> DatasetReference:
        return self.client.dataset(
            dataset_id=dataset_name,
            project=project_id or self.project_id,
        )

    def _to_sql_format(self, value: Any) -> Any:
        if isinstance(value, str):
            return f"'{value}'"
        elif isinstance(value, datetime):
            return f"'{self.date_to_str(value)}'"
        return value

    def _get_type(self, variable):
        TYPES = {
            'int': 'INT64',
            'str': 'STRING',
            'datetime': 'DATETIME',
            'date': 'DATE',
            'bool': 'BOOL',
            'float': 'FLOAT64',
            'Decimal': 'FLOAT64',
        }

        python_type = type(variable).__name__
        param_type = TYPES.get(python_type, None)
        if param_type is None:
            raise Exception('Parameter with type {param_type} not supported')
        return param_type

    def _get_value(self, variable):
        TYPES = {
            'Decimal': float,
        }
        python_type = type(variable).__name__
        param_type = TYPES.get(python_type, None)
        if param_type is not None:
            return param_type(variable)
        return variable

    def _create_param(self, key, value):
        if isinstance(value, list):
            param_type = self._get_type(variable=value[0])
            param_class = bigquery.ArrayQueryParameter
        else:
            param_type = self._get_type(variable=value)
            param_class = bigquery.ScalarQueryParameter
        return param_class(key, param_type, self._get_value(value))

    def execute(
            self,
            sql,
            params=None,
            legacy_sql=False,
            job_id=None,
            destination_table_name=None,
            destination_dataset_name=None,
            destination_project=None,
            write_disposition=None
    ):
        job_config = bigquery.QueryJobConfig()
        if destination_table_name or destination_dataset_name:
            if not destination_table_name and destination_dataset_name:
                raise exceptions.ValidationError(
                    f"Both destination_dataset_name and destination_table_name must be provided."
                )
            destination_project = destination_project or self.project_id
            destination_dataset = destination_dataset_name
            destination = f'{destination_project}.{destination_dataset}.{destination_table_name}'
            job_config.destination = destination
            job_config.write_disposition = write_disposition or 'WRITE_EMPTY'

        if params:
            query_params = [
                self._create_param(key, value) for key, value in params.items()
            ]
            job_config.query_parameters = query_params
        job_config.use_legacy_sql = legacy_sql

        query_job = self.client.query(sql, job_id=job_id, job_config=job_config)
        return self._wait_for_job(job=query_job)

    def insert_rows(self, dataset_name: str, table_name: str, rows, project_id: str = None):
        table = self.get_table(
            table_name=table_name,
            project_id=project_id,
            dataset_name=dataset_name,
        )
        errors = self.client.insert_rows(table=table, rows=rows)
        if errors:
            raise Exception(f'Bigquery insert error: {errors}')

    def get_table(self, table_name, project_id=None, dataset_name=None) -> Table:
        dataset_ref = self._dataset_ref(project_id=project_id, dataset_name=dataset_name)
        table_ref = dataset_ref.table(table_name)
        return self.client.get_table(table_ref)

    def load(
            self,
            table_name,
            gcs_file,
            project_id=None,
            dataset_name=None,
            schema=None,
            sync=False,
            truncate=None,
    ):
        job_config = bigquery.LoadJobConfig()
        if schema:
            job_config.schema = schema
        else:
            job_config.autodetect = True
            job_config._properties['load']['schemaUpdateOptions'] = ['ALLOW_FIELD_ADDITION']

        extension = gcs_file.split('.')[-1]
        if extension == 'json':
            source_format = bigquery.job.SourceFormat.NEWLINE_DELIMITED_JSON
        elif extension == 'csv':
            source_format = bigquery.job.SourceFormat.CSV
        else:
            message = 'Unsupported BigQuery Source format {}'.format(extension)
            raise exceptions.UnsupportedFormatException(message)

        job_config.source_format = source_format
        if truncate:
            job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
        elif truncate is False:
            job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND

        dataset_ref = self._dataset_ref(
            project_id=project_id or self.project_id,
            dataset_name=dataset_name,
        )
        target_ref = dataset_ref.table(table_name)
        load_job = self.client.load_table_from_uri(
            source_uris=gcs_file,
            destination=target_ref,
            job_config=job_config,
        )

        if sync:
            self._wait_for_job(job=load_job)

    def copy(
            self,
            source_table_name,
            destination_table_name,
            destination_dataset_name,
            destination_project=None,
            sync=False,
            **kwargs,
    ):
        source_ref = self.get_table(table_name=source_table_name)
        dataset_ref = self._dataset_ref(
            project_id=destination_project or self.project_id,
            dataset_name=destination_dataset_name,
        )
        target_ref = dataset_ref.table(destination_table_name)

        job = self.client.copy_table(source_ref, target_ref, **kwargs)
        if sync:
            self._wait_for_job(job=job)

    @classmethod
    def date_to_str(cls, dt, table_suffix=False):
        if table_suffix:
            mask = '%Y%m%d'
        elif isinstance(dt, datetime):
            mask = '%Y-%m-%d %H:%M:%S'
        else:
            mask = '%Y-%m-%d'
        return dt.strftime(mask)

    def _wait_for_job(self, job):
        try:
            return job.result()
        except Exception as e:
            raise exceptions.BigQueryJobError(job) from e



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gcp_pilot']

package_data = \
{'': ['*']}

install_requires = \
['google-api-python-client>=1,<2']

extras_require = \
{'bigquery': ['google-cloud-bigquery>=1,<2'],
 'build': ['google-cloud-build>=3,<4'],
 'datastore': ['google-cloud-datastore>=2,<3'],
 'pubsub': ['google-cloud-pubsub>=2,<3'],
 'sheets': ['gspread>=3,<4'],
 'speech': ['google-cloud-speech>=1,<2'],
 'storage': ['google-cloud-storage>=1,<2'],
 'tasks': ['google-cloud-tasks>=2,<3', 'google-cloud-scheduler>=2,<3']}

setup_kwargs = {
    'name': 'gcp-pilot',
    'version': '0.7.1',
    'description': 'Google Cloud Platform Friendly Pilot',
    'long_description': None,
    'author': 'Joao Daher',
    'author_email': 'joao@daher.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
